from pathlib import Path

import mrcfile
import numpy as np

from yet_another_imod_wrapper import align_tilt_series_using_patch_tracking

TEST_DATA_DIR = Path(__file__).parent.parent / 'test_data'

align_tilt_series_using_patch_tracking(
    tilt_series=mrcfile.read(TEST_DATA_DIR / 'TS_01.mrc'),
    tilt_angles=np.arange(-60, 63, 3),
    pixel_size=1.35,
    nominal_rotation_angle=85,
    patch_size=1000,
    patch_overlap_percentage=33,
    basename='TS_01',
    output_directory=Path('test_TS_01_patch_tracking')
)



from pathlib import Path

import pytest


@pytest.fixture
def test_data_directory() -> Path:
    return Path(__file__).parent / 'test_data'

@pytest.fixture
def align_log_file(test_data_directory) -> Path:
    return test_data_directory / 'align.log'


import numpy as np
import mrcfile

from yet_another_imod_wrapper import utils


def test_find_optimal_power_of_2_binning_factor():
    result = utils.binning.find_optimal_power_of_2_binning_factor(
        src_pixel_size=2, target_pixel_size=10
    )
    assert result == 4


def test_find_optimal_integer_binning_factor():
    result = utils.binning.find_optimal_integer_binning_factor(
        src_pixel_size=3, target_pixel_size=10,
    )
    assert result == 3


def test_prepare_etomo_directory(tmp_path):
    directory = tmp_path / 'imod'
    basename = 'TS'
    tilt_series = np.arange(41 * 100).reshape((41, 10, 10))
    tilt_angles = np.arange(-60, 63, 3)
    etomo_directory = utils.etomo.prepare_etomo_directory(
        tilt_series=tilt_series,
        tilt_angles=tilt_angles,
        basename=basename,
        directory=directory
    )

    assert etomo_directory.tilt_series_file.exists()
    output_tilt_series = mrcfile.read(etomo_directory.tilt_series_file)
    assert np.allclose(output_tilt_series, tilt_series)

    assert etomo_directory.rawtlt_file.exists()
    output_tilt_angles = np.loadtxt(etomo_directory.rawtlt_file)
    assert np.allclose(tilt_angles, output_tilt_angles)


def test_get_batchruntomo_command(tmp_path):
    basename = 'base'
    directive_file = tmp_path / 'directive'
    result = utils.etomo._get_batchruntomo_command(
        directory=tmp_path,
        basename=basename,
        directive_file=directive_file
    )
    assert result == [
        'batchruntomo',
        '-DirectiveFile', f'{directive_file}',
        '-CurrentLocation', f'{tmp_path}',
        '-RootName', basename,
        '-EndingStep', '6'
    ]


def test_get_tilt_angle_offset(align_log_file):
    """Test getting tilt angle offset from align.log."""
    result = utils.etomo.get_tilt_angle_offset(align_log_file)
    assert result == 10.02


def test_get_input_rotation_angle(align_log_file):
    """Test getting an initial tilt-angle from align.log file."""
    result = utils.etomo.get_input_tilt_axis_rotation_angle(align_log_file)
    assert result == 85



import os
from os import PathLike
from typing import Dict

import numpy as np


def read_tlt(file: os.PathLike) -> np.ndarray:
    """Read an IMOD tlt file into an (n, ) numpy array."""
    return np.loadtxt(fname=file, dtype=float).reshape(-1)


def read_xf(file: os.PathLike) -> np.ndarray:
    """Read an IMOD xf file into an (n, 6) numpy array.

    The xf file with alignment transforms contains one
    line per view, each with a linear transformation specified by six numbers:
        A11 A12 A21 A22 DX DY
    where the coordinate (X, Y) is transformed to (X', Y') by:
        X' = A11 * X + A12 * Y + DX
        Y' = A21 * X + A22 * Y + DY
    """
    return np.loadtxt(fname=file, dtype=float).reshape((-1, 6))


def read_adoc(adoc_file: PathLike) -> Dict[str, str]:
    """Read an IMOD adoc file into a dictionary."""
    with open(adoc_file) as adoc:
        lines = adoc.readlines()
    lines = [
        line.strip().split('=')
        for line in lines
        if not line.startswith('#')
    ]
    lines = [line for line in lines if len(line) == 2]
    return {k.strip(): v.strip() for k, v in lines}


def write_adoc(data: dict, output_filename: PathLike):
    """Write an IMOD adoc file from a dictionary."""
    with open(output_filename, 'w') as adoc:
        for k, v in data.items():
            adoc.write(f"{k} = {v}\n")



from os import PathLike
from pathlib import Path
from typing import Dict, Any, Tuple, Sequence

import numpy as np

from .utils.io import read_adoc
from .constants import TARGET_PIXEL_SIZE_FOR_ALIGNMENT, BATCHRUNTOMO_CONFIG_PATCH_TRACKING
from .utils.installation import check_imod_installation
from .utils.binning import find_optimal_power_of_2_binning_factor
from .utils.etomo import prepare_etomo_directory, run_batchruntomo, EtomoOutput


def align_tilt_series_using_patch_tracking(
        tilt_series: np.ndarray,
        tilt_angles: Sequence[float],
        nominal_rotation_angle: float,
        pixel_size: float,
        patch_size: float,
        patch_overlap_percentage: float,
        basename: str,
        output_directory: Path,
        skip_if_completed: bool = False
) -> EtomoOutput:
    """Run patch-tracking alignment in IMOD on a single tilt-series.

    Parameters
    ----------
    tilt_series: (n, y, x) array of 2D tilt-images in a tilt-series.
    tilt_angles: nominal stage tilt-angles from the microscope.
    pixel_size: pixel size of the tilt-series in angstroms-per-pixel
    nominal_rotation_angle: initial estimate for the rotation angle of the tilt
        axis. https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    patch_size: sidelength of patches to be tracked in angstroms.
    patch_overlap_percentage: overlap between patches in each direction.
        e.g. 33 for 33% overlap in each direction.
    basename: basename for IMOD files.
    skip_if_completed: skip alignment if previous results found.
    output_directory: tilt-series directory for IMOD.
    """
    check_imod_installation()
    output_directory = Path(output_directory)
    etomo_output = prepare_etomo_directory(
        directory=output_directory,
        tilt_series=tilt_series,
        tilt_angles=tilt_angles,
        basename=basename,
    )
    patch_size_px = int(patch_size / pixel_size)
    directive = generate_patch_tracking_alignment_directive(
        tilt_series_file=etomo_output.tilt_series_file,
        pixel_size=pixel_size,
        rotation_angle=nominal_rotation_angle,
        patch_size_xy=(patch_size_px, patch_size_px),
        patch_overlap_percentage=patch_overlap_percentage,
    )
    if etomo_output.contains_alignment_results is False or skip_if_completed is False:
        run_batchruntomo(
            directory=output_directory,
            basename=basename,
            directive=directive,
        )
        if etomo_output.contains_alignment_results is False:
            raise RuntimeError(f'{basename} failed to align correctly.')
    return etomo_output


def generate_patch_tracking_alignment_directive(
        tilt_series_file: PathLike,
        pixel_size: float,
        rotation_angle: float,
        patch_size_xy: Tuple[int, int],
        patch_overlap_percentage: float,
) -> Dict[str, Any]:
    """Generate a fiducial-based alignment directive file for batchruntomo

    Parameters
    ----------
    tilt_series_file : file containing the tilt-series stack.
    pixel_size : pixel size in the tilt-series in Angstroms per pixel.
    fiducial_size : approximate fiducial size in nanometers.
    rotation_angle : initial estimate for the rotation angle
        https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    patch_size_xy: patch size in the unbinned tilt-series
    patch_overlap_percentage: overlap between patches in each direction.
        e.g. 33 for 33% overlap in each direction.
    """
    alignment_binning_factor = find_optimal_power_of_2_binning_factor(
        src_pixel_size=pixel_size, target_pixel_size=TARGET_PIXEL_SIZE_FOR_ALIGNMENT
    )
    patch_size_xy_binned = [int(s / alignment_binning_factor) for s in patch_size_xy]
    patch_overlap_factor = patch_overlap_percentage / 100

    directive = read_adoc(BATCHRUNTOMO_CONFIG_PATCH_TRACKING)
    directive['setupset.copyarg.stackext'] = Path(tilt_series_file).suffix
    directive['setupset.copyarg.rotation'] = str(rotation_angle)
    directive['setupset.copyarg.pixel'] = str(pixel_size / 10)
    directive['comparam.prenewst.newstack.BinByFactor'] = str(alignment_binning_factor)
    directive['comparam.xcorr_pt.tiltxcorr.OverlapOfPatchesXandY'] = f'{patch_overlap_factor},' \
                                                                     f'{patch_overlap_factor}'
    directive['comparam.xcorr_pt.tiltxcorr.SizeOfPatchesXandY'] = f'{patch_size_xy_binned[0]},' \
                                                                  f'{patch_size_xy_binned[1]}'

    return directive



import numpy as np


def find_optimal_integer_binning_factor(
        src_pixel_size: float, target_pixel_size: float
) -> int:
    binning_factors = np.arange(1, 30)
    return _find_optimal_binning_factor(binning_factors, src_pixel_size, target_pixel_size)


def find_optimal_power_of_2_binning_factor(
        src_pixel_size: float, target_pixel_size: float
) -> int:
    binning_factors = 2 ** np.arange(6)
    return _find_optimal_binning_factor(binning_factors, src_pixel_size, target_pixel_size)


def _find_optimal_binning_factor(
        binning_factors: np.ndarray,
        src_pixel_size: float,
        target_pixel_size: float
) -> int:
    binned_pixel_sizes = binning_factors * src_pixel_size
    pixel_size_deltas = np.abs(binned_pixel_sizes - target_pixel_size)
    return binning_factors[np.argmin(pixel_size_deltas)]



from pathlib import Path

import numpy as np
import mrcfile

from yet_another_imod_wrapper import align_tilt_series_using_fiducials

TEST_DATA_DIR = Path(__file__).parent.parent / 'test_data'

align_tilt_series_using_fiducials(
    tilt_series=mrcfile.read(TEST_DATA_DIR / 'TS_01.mrc'),
    tilt_angles=np.arange(-60, 63, 3),
    pixel_size=1.35,
    fiducial_size=10,
    nominal_rotation_angle=85,
    basename='TS_01',
    output_directory=Path('test_TS_01_fiducials')
)



import os
import shutil
from pathlib import Path

from packaging import version

from yet_another_imod_wrapper.constants import MINIMUM_IMOD_VERSION


def check_imod_installation() -> None:
    """Check that IMOD is installed and satisfies the minimum version requirements."""
    if not _imod_is_installed():
        raise RuntimeError('No IMOD installation found.')
    imod_version = _get_imod_version()
    if imod_version < version.parse(MINIMUM_IMOD_VERSION):
        raise RuntimeError(f'Ensure IMOD version {MINIMUM_IMOD_VERSION} or higher is installed. '
                           f'Your version is {imod_version}')


def _imod_is_installed() -> bool:
    """Check if IMOD is installed and on the PATH."""
    return shutil.which('imod') is not None


def _get_imod_directory() -> Path:
    """Get path to IMOD installation."""
    imod_dir = os.environ.get('IMOD_DIR')
    if imod_dir is None:
        raise ValueError('IMOD_DIR is not set, please check your IMOD installation.')
    return Path(imod_dir)


def _get_imod_version() -> version.Version:
    """Get IMOD version."""
    imod_dir = _get_imod_directory()
    with open(imod_dir / 'VERSION') as f:
        return version.parse(f.readline().strip())



import subprocess
import tempfile
from pathlib import Path
from typing import Sequence, List, Dict, Union

import mrcfile
import numpy as np

from .io import write_adoc


class EtomoOutput:
    def __init__(self, basename: str, directory: Path):
        self.directory: Path = directory
        self.basename: str = basename

    @property
    def tilt_series_file(self) -> Path:
        return self.directory / f'{self.basename}.mrc'

    @property
    def rawtlt_file(self) -> Path:
        return self.directory / f'{self.basename}.rawtlt'

    @property
    def xf_file(self) -> Path:
        return self.directory / f'{self.basename}.xf'

    @property
    def tlt_file(self) -> Path:
        return self.directory / f'{self.basename}.tlt'

    @property
    def edf_file(self) -> Path:
        return self.directory / f'{self.basename}.edf'

    @property
    def align_log_file(self) -> Path:
        return self.directory / 'align.log'

    @property
    def is_ready_for_alignment(self) -> bool:
        return self.tilt_series_file.exists() and self.rawtlt_file.exists()

    @property
    def contains_alignment_results(self) -> bool:
        return self.xf_file.exists() and self.tlt_file.exists()


def prepare_etomo_directory(
        directory: Path,
        tilt_series: np.ndarray,
        tilt_angles: Sequence[float],
        basename: str,
) -> EtomoOutput:
    """Prepare a directory for IMOD tilt-series alignment."""
    directory.mkdir(exist_ok=True, parents=True)
    output = EtomoOutput(basename=basename, directory=directory)
    tilt_series_file = output.tilt_series_file
    data_on_disk_shape = None
    if tilt_series_file.exists():
        with mrcfile.open(tilt_series_file, header_only=True) as mrc:
            data_on_disk_shape = (mrc.header.nz, mrc.header.ny, mrc.header.nx)
    if not np.array_equal(tilt_series.shape, data_on_disk_shape):
        mrcfile.write(
            tilt_series_file,
            tilt_series.astype(np.float32),
            overwrite=True
        )
    np.savetxt(output.rawtlt_file, tilt_angles, fmt='%.2f', delimiter='')
    return output


def run_batchruntomo(
        directory: Path, basename: str, directive: Dict[str, str]
) -> None:
    """Run batchruntomo on a single tilt-series with a specified directive."""
    with tempfile.TemporaryDirectory() as temporary_directory:
        directive_file = Path(temporary_directory) / 'directive.adoc'
        write_adoc(directive, directive_file)
        batchruntomo_command = _get_batchruntomo_command(
            directory=directory,
            basename=basename,
            directive_file=directive_file
        )
        with open(directory / 'log.txt', mode='w') as log:
            subprocess.run(batchruntomo_command, stdout=log, stderr=log)


def _get_batchruntomo_command(
        directory: Path, basename: str, directive_file: Path
) -> List[str]:
    """Get batchruntomo command."""
    command = [
        'batchruntomo',
        '-DirectiveFile', f'{directive_file}',
        '-CurrentLocation', f'{directory}',
        '-RootName', basename,
        '-EndingStep', '6'
    ]
    return command


def get_tilt_angle_offset(align_log_file: Path) -> Union[float, None]:
    """Get the total tilt angle offset from an align.log file."""
    with open(align_log_file, mode='r') as file:
        for line in file:
            if 'Total tilt angle change =' in line:
                return float(line.strip().split('=')[-1])
    return None


def get_input_tilt_axis_rotation_angle(align_log_file: Path) -> Union[float, None]:
    """Get the input rotation angle from an align.log file.

    This is useful when determining the absolute rotation angle
    from the rotation matrix in a .xf file.
    """
    with open(align_log_file, mode='r') as file:
        for line in file:
            if 'RotationAngle =' in line:
                return float(line.strip().split('=')[-1])
